/*
 *  Copyright (C) 2002-2004  Massachusetts Institute of Technology
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 *  USA
 */

#include "stdinc.h"
#include "pastdbfe.h"

#include "cfsw_record.h"
#include "snapshot.h"
#include "ranges.h"

int
cfsr::init ()
{
  return 0;
}

//
// lookup
//

struct lookup_obj {
  bigint dir;
  str name;
};

static void 
lookup_cb3 (bigint dir, str name,
	    callback<void, bigint, cfsrw_status>::ref cb,
            ptr<cfs_data> data)
{
  if (!data || data->type != CFS_INODE) {
    cb (0, CFSRWERR_IO);
    return;
  }

  bigint object;
  cfs_inode *inode = data->inode;
  if (inode->type == CFSLNK)
    object = inode->lnk->handle;
  else
    object = inode->reg->handle;
  cb (object, CFSRW_OK);
}

static void
lookup_cb2 (ptr<snapshot> s, ptr<cfs_data> d, uint32 blocknum,
            ptr<lookup_obj> object,
	    callback<void, bigint, cfsrw_status>::ref cb,
	    ptr<cfs_data> data)
{
  cfs_inode *inode = d->inode;
  
  if (!data || data->type != CFS_DIRBLK) {
    cb (0, CFSRWERR_IO);
    return;
  }

  cfs_dirent *e = (*(*data).dir).entries;
  while (e && (e->name != object->name))
    e = e->nextentry;
    
  if (!e) {
    if (inode->reg->direct.size () > blocknum+1) {
      blocknum++;
      bigint dirblock = inode->reg->direct[blocknum];
      ns_retrieve_cfs_readonly
        (dirblock, wrap (lookup_cb2, s, d, blocknum, object, cb));
      return;
    }
    cb (0, CFSRWERR_NOENT);
  }
  else {
    ptr<dbrec> q = bigint2dbrec (e->fh);
    ptr<dbrec> v = s->fmap_db->lookup (q);
    if (!v) {
      cb (0, CFSRWERR_IO);
      return;
    }
    ns_retrieve_cfs_readonly
      (dbrec2bigint(v), wrap (lookup_cb3, object->dir, object->name, cb)); 
  }
}

static void
lookup_cb1 (ptr<snapshot> s, bigint dir, str name,
            callback<void, bigint, cfsrw_status>::ref cb,
	    ptr<cfs_data> data)
{
  if (!data || data->type != CFS_INODE || data->inode->type != CFSDIR) {
    cb (0, CFSRWERR_IO);
    return;
  }

  cfs_inode *inode = data->inode;
  
  if (inode->reg->direct.size() > 0) {
    bigint dirblock = inode->reg->direct[0];
    ptr<lookup_obj> object = New refcounted<lookup_obj>;
    object->dir = dir;
    object->name = name;
    ns_retrieve_cfs_readonly
      (dirblock, wrap (lookup_cb2, s, data, 0, object, cb));
  }
  else {
    cb (0, CFSRWERR_NOENT);
    return;
  }
}

void 
cfsr::lookup (sptr s, bigint dir, str name,
              callback<void, bigint, cfsrw_status>::ref cb)
{
  ptr<dbrec> q = bigint2dbrec (dir);
  ptr<dbrec> v = obj2impl (s)->fmap_db->lookup (q);
  
  if (!v) {
    cb (0, CFSRWERR_NOENT);
    return;
  }
  ns_retrieve_cfs_readonly
    (dbrec2bigint(v), wrap (lookup_cb1, obj2impl (s), dir, name, cb));
}

//
// getattr
//

static void 
getattr_cb (bigint object, callback<void, ptr<cfsr_attr>, cfsrw_status>::ref cb,
            ptr<cfs_data> data)
{
  if (!data || data->type != CFS_INODE) {
    cb ((ptr<cfsr_attr>) 0, CFSRWERR_IO);
    return;
  }
  
  cfs_inode *inode = data->inode;
  ptr<cfsr_attr> a = New refcounted<cfsr_attr>;
  a->type = inode->type;

  if (inode->type == CFSLNK) {
    a->nlink = inode->lnk->attr.nlink;
    a->size = inode->lnk->attr.size;
    a->mode = inode->lnk->attr.mode;
    a->revision = inode->lnk->attr.revision;
    a->commit = inode->lnk->attr.commit;
    a->mode_commit = inode->lnk->attr.mode_commit;
    cb (a, CFSRW_OK);
  }
  else {
    a->nlink = inode->reg->attr.nlink;
    a->size = inode->reg->attr.size;
    a->mode = inode->reg->attr.mode;
    a->revision = inode->reg->attr.revision;
    a->commit = inode->reg->attr.commit;
    a->mode_commit = inode->reg->attr.mode_commit;
    cb (a, CFSRW_OK);
  }
}

void 
cfsr::getattr (sptr s, bigint object,
               callback<void, ptr<cfsr_attr>, cfsrw_status>::ref cb)
{
  ptr<dbrec> q = bigint2dbrec (object);
  ptr<dbrec> v = obj2impl (s)->fmap_db->lookup (q);
  
  if (!v) {
    cb ((ptr<cfsr_attr>) 0, CFSRWERR_NOENT);
    return;
  }
  ns_retrieve_cfs_readonly
    (dbrec2bigint(v), wrap (getattr_cb, object, cb));
}

//
// findver
//

static void
parse_chain (ptr<ns_block> b, cfs_vchain &c)
{
  xin x (str (b->data, b->len));
  int n;
  x.rint (n);
  c.v.revision = (unsigned) n;
  x.rbigint (c.v.logid);
  x.rbigint (c.v.commit);
  x.rbigint (c.v.saddr);
  for (unsigned i=0; i<CFS_VCN; i++) {
    x.rint (n);
    c.ptrs[i] = (unsigned) n;
    x.rbigint (c.paddr[i]);
  }
}

static void
findver_chain_cb (unsigned rev,
                  callback<void, cfsr_findver_res, cfsrw_status>::ref cb,
		  ptr<ns_block> b)
{
  if (!b) {
    cb (cfsr_findver_res (0, 0), CFSRWERR_IO);
    return;
  }

  cfs_vchain c;
  parse_chain (b, c);
  
  if (c.v.revision == rev) {
    cb (cfsr_findver_res (c.v.logid, c.v.commit), CFSRW_OK);
    return;
  }
  else {
    unsigned i=0;
    for (i=0; i<CFS_VCN; i++)
      if (c.ptrs[i] <= rev)
	break;
    if (i == CFS_VCN || c.ptrs[i] != rev)
      i--;
    if (c.ptrs [i] == 0) {
      cb (cfsr_findver_res (0, 0), CFSRWERR_NOREVISION);
      return;
    }
    else
      findver_chain_cb (rev, cb, nsobj->retrieve_local (c.paddr[i]));
  }
}

static void 
findver_cb (sptr s, bigint object, unsigned rev,
            callback<void, cfsr_findver_res, cfsrw_status>::ref cb,
            ptr<cfs_data> data)
{
  if (!data || data->type != CFS_INODE) {
    cb (cfsr_findver_res (0, 0), CFSRWERR_IO);
    return;
  }
  
  cfs_inode *inode = data->inode;
  cfs_history *history;
  cfs_attr *attr;

  if (inode->type == CFSLNK) {
    history = &inode->lnk->history;
    attr = &inode->lnk->attr;
  }
  else {
    history = &inode->reg->history;
    attr = &inode->reg->attr;
  }
 
  if (attr->revision == rev) {
    // current version is first in version history
    bigint l = history->versions [0].logid; 
    bigint cr = s->branch().rptr;
    cb (cfsr_findver_res (l, cr), CFSRW_OK);
    return;
  }

  for (unsigned i = 1; i < history->versions.size (); i++) {
    if (history->versions [i].revision == rev) {
      cb (cfsr_findver_res (history->versions [i].logid,
	                    history->versions [i].commit), CFSRW_OK);
      return;
    }
  }

  if (history->chain.v.revision == 0) {
    cb (cfsr_findver_res (0, 0), CFSRWERR_NOREVISION);
    return;
  }

  else if (history->chain.v.revision == rev) {
    cb (cfsr_findver_res (history->chain.v.logid,
	                  history->chain.v.commit), CFSRW_OK);
    return;
  }

  else {
    unsigned i=0;
    for (i=0; i<CFS_VCN; i++)
      if (history->chain.ptrs[i] <= rev)
	break;
    if (i == CFS_VCN || history->chain.ptrs[i] != rev)
      i--;
    if (history->chain.ptrs [i] == 0) {
      cb (cfsr_findver_res (0, 0), CFSRWERR_NOREVISION);
      return;
    }
    else
      findver_chain_cb
	(rev, cb, nsobj->retrieve_local (history->chain.paddr[i]));
  }
}

void
cfsr::findver (sptr s, bigint object, unsigned rev,
               callback<void, cfsr_findver_res, cfsrw_status>::ref cb)
{
  ptr<dbrec> q = bigint2dbrec (object);
  ptr<dbrec> v = obj2impl (s)->fmap_db->lookup (q);
  
  if (!v) {
    cb (cfsr_findver_res (0, 0), CFSRWERR_NOENT);
    return;
  }
  ns_retrieve_cfs_readonly
    (dbrec2bigint(v), wrap (findver_cb, s, object, rev, cb));
}

//
// readdir
//

struct readdir_obj {
  uint64 cookie;
  uint32 count;
  uint32 n;
  rpc_ptr<cfs_rpc_dirent> *direntp;
  ptr<cfs_dirlist> res;
};

static inline bool 
dup_entry (rpc_ptr<cfs_rpc_dirent> *dep, cfs_dirent *e)
{
  if (!dep) 
    return false;
  rpc_ptr<cfs_rpc_dirent> de = *dep;
  while (de && de->name != e->name)
    de = de->nextentry;
  if (de)
    return true;
  return false;
}

static void
got_dir_block (bigint dir, ptr<cfs_data> d, uint32 blocknum,
               ptr<readdir_obj> obj,
               callback<void, ptr<cfs_dirlist>, cfsrw_status>::ref cb,
	       ptr<cfs_data> data)
{
  if (!data || data->type != CFS_DIRBLK) {
    cb ((ptr<cfs_dirlist>) 0, CFSRWERR_IO);
    return;
  }
  
  cfs_dirent *e = data->dir->entries;
  while (e) {
    // check if there's enough space in RPC reply, &c. this avoids a
    // crash when rmdir checks if a directory is empty, and also
    // perhaps when a directory is too large to fit in a single
    // READDIR reply.

    if (!dup_entry (&(obj->res->entries), e)) {
      if (obj->cookie && obj->n < obj->cookie)
	obj->n++;
      else {
	if (obj->count < (e->name.len ()+30)) {
	  cb (obj->res, CFSRW_OK);
	  return;
	}
	  
        obj->n++;
	obj->count -= (e->name.len ()+30);
        (*obj->direntp).alloc();
	(*obj->direntp)->fh = e->fh;
        (*obj->direntp)->name = e->name;
        (*obj->direntp)->cookie = obj->n;
        obj->direntp = &(*obj->direntp)->nextentry;
      }
    }
    e = e->nextentry;
  }

  cfs_inode *inode = d->inode;
  if (inode->reg->direct.size() > blocknum+1) {
    blocknum++;
    bigint dirblock = inode->reg->direct[blocknum];
    ns_retrieve_cfs_readonly
      (dirblock, wrap (got_dir_block, dir, d, blocknum, obj, cb));
    return;
  }

  obj->res->eof = true;
  cb (obj->res, CFSRW_OK);
}

static void 
readdir_got_dir (bigint dir, ptr<readdir_obj> obj,
               callback<void, ptr<cfs_dirlist>, cfsrw_status>::ref cb,
	       ptr<cfs_data> data)
{
  if (!data || data->type != CFS_INODE) {
    cb ((ptr<cfs_dirlist>) 0, CFSRWERR_IO);
    return;
  }
  
  cfs_inode *inode = data->inode;
  if (inode->type != CFSDIR) {
    cb ((ptr<cfs_dirlist>) 0, CFSRWERR_NOTDIR);
    return;
  }

  if (inode->reg->direct.size() > 0) {
    bigint dirblock = inode->reg->direct[0];
    ns_retrieve_cfs_readonly
      (dirblock, wrap (got_dir_block, dir, data, 0, obj, cb));
  }
  else {
    obj->res->eof = true;
    cb (obj->res, CFSRW_OK);
    return;
  }
}

void
cfsr::readdir (sptr s, bigint dir, uint64 cookie, uint32 count,
               callback<void, ptr<cfs_dirlist>, cfsrw_status>::ref cb)
{
  ptr<dbrec> q = bigint2dbrec (dir);
  ptr<dbrec> v = obj2impl (s)->fmap_db->lookup (q);
  if (!v) {
    cb ((ptr<cfs_dirlist>)0, CFSRWERR_NOENT);
    return;
  }
  ptr<readdir_obj> obj = New refcounted<readdir_obj>;
  obj->cookie = cookie;
  obj->count = count;
  obj->n = 0;
  obj->res = New refcounted<cfs_dirlist>;
  obj->res->eof = false;
  obj->direntp = &(obj->res->entries);

  ns_retrieve_cfs_readonly
    (dbrec2bigint(v), wrap(readdir_got_dir, dir, obj, cb));
}

//
// readlink
//

static void 
readlink_got_file (bigint file, callback<void, str, cfsrw_status>::ref cb,
	           ptr<cfs_data> data)
{
  if (!data || data->type != CFS_INODE) {
    str s;
    cb (s, CFSRWERR_IO);
    return;
  }
  
  cfs_inode *inode = data->inode;
  if (inode->type != CFSLNK) {
    str s;
    cb (s, CFSRWERR_NOTLINK);
    return;
  }
  cb (inode->lnk->dest, CFSRW_OK);
}

void
cfsr::readlink (sptr s, bigint file,
                callback<void, str, cfsrw_status>::ref cb)
{
  ptr<dbrec> q = bigint2dbrec (file);
  ptr<dbrec> v = obj2impl (s)->fmap_db->lookup (q);
  if (!v) {
    str s;
    cb (s, CFSRWERR_NOENT);
    return;
  }
  ns_retrieve_cfs_readonly
    (dbrec2bigint(v), wrap(readlink_got_file, file, cb));
}


//
// read
//

struct read_obj {
  ptr<snapshot> s;
  int outstanding_fetches;
  int fetch_errors;
 
  char *buf;
  uint32 len;
  ranges rng;

  uint64 offset;
  uint32 count;
  callback<void, ptr<ns_block>, bool, cfsrw_status>::ref cb;

  read_obj (ptr<snapshot> s,
            callback<void, ptr<ns_block>, bool, cfsrw_status>::ref cb)
    : s(s), outstanding_fetches(0), fetch_errors(0), buf(0), cb(cb) {}
  
  ~read_obj () {
    if (buf)
      delete buf;
  }
};

static void
done_fill_buf (ptr<read_obj> obj)
{
  uint64 c = obj->rng.end () - obj->offset;
  if (c > obj->count)
    c = obj->count;
  bool eof = ((obj->offset+c) >= obj->rng.end ());
  ptr<ns_block> b = New refcounted<ns_block> (obj->buf, c);
  obj->cb (b, eof, CFSRW_OK);
}

static void
check_fill (ptr<read_obj> obj)
{
  if (obj->outstanding_fetches == 0) {
    // don't need to add 1 to _count here, since we know the
    // size of the file in snapshot
    if (obj->rng.check (obj->offset, obj->count) || 
	obj->fetch_errors == 0)
      done_fill_buf (obj);
    else
      obj->cb ((ptr<ns_block>) 0, true, CFSRWERR_IO);
  }
}

static void 
got_file_block (ptr<read_obj> obj, uint64 gs, uint64 gl,
                ptr<cfs_data> data)
{
  obj->outstanding_fetches--;
  if (!data || data->type != CFS_FILEBLK) {
    obj->fetch_errors++;
    check_fill (obj);
    return;
  }

  // uint64 blocknum = gs / SN_BLKSIZE;
  // uint64 blockstart = blocknum * SN_BLKSIZE;
  //
  // XXX the following logic seems broken... we didn't load
  // up the entire "gl" number of bytes
  //
  uint64 start = (gs % SN_BLKSIZE);
  uint64 size = obj->offset + obj->count - gs;
  if (size > gl)
    size = gl;
  if (size > SN_BLKSIZE - start)
    size = SN_BLKSIZE - start;

  uint64 bi = gs - obj->offset;
  uint64 wi = start;

  memcpy (obj->buf + bi, data->data->base () + wi, size);

  obj->rng.add (gs, size);

  check_fill (obj);
}

static void 
got_sindir_block (ptr<read_obj> obj, uint32 blocknum,
                  uint64 gs, uint64 gl, ptr<cfs_data> data)
{
  obj->outstanding_fetches--;
  if (!data || data->type != CFS_INDIR) {
    obj->fetch_errors++;
    check_fill (obj);
    return;
  }
  bigint block = data->indir->handles[blocknum];
  obj->outstanding_fetches++;
  ns_retrieve_cfs_readonly
    (block, wrap(got_file_block, obj, gs, gl));
}

static void 
got_dindir_block (ptr<read_obj> obj, uint32 blocknum, uint32 blockoffst,
                  uint64 gs, uint64 gl, ptr<cfs_data> data)
{
  obj->outstanding_fetches--;
  if (!data || data->type != CFS_INDIR) {
    obj->fetch_errors++;
    check_fill (obj);
    return;    
  }
  bigint block = data->indir->handles[blocknum];
  obj->outstanding_fetches++;
  ns_retrieve_cfs_readonly
    (block, wrap (got_sindir_block, obj, blockoffst, gs, gl));
}

static void 
got_tindir_block (ptr<read_obj> obj, uint32 tblocknum, uint32 dblocknum,
                  uint64 gs, uint64 gl, ptr<cfs_data> data)
{
  obj->outstanding_fetches--;
  if (!data || data->type != CFS_INDIR) {
    obj->fetch_errors++;
    check_fill (obj);
    return;    
  }
  bigint block = data->indir->handles[dblocknum];
  obj->outstanding_fetches++;
  ns_retrieve_cfs_readonly
    (block,
     wrap (got_dindir_block, obj, dblocknum, dblocknum % SN_NFH, gs, gl));
}

static void
fill_buf (ptr<read_obj> obj, ptr<cfs_data> d, uint64 gs, uint64 gl)
{
  cfs_inode_reg *rnode = d->inode->reg;

  // fill buf with data starting from gs to gl or end of block, which
  // ever is smaller. commit to ranges and return.
  
  if (gs >= rnode->attr.size)
    return;

  bigint n;
  uint32 blocknum = gs / SN_BLKSIZE;
  if (blocknum < CFS_NDIR) {
    n = rnode->direct[blocknum];
    obj->outstanding_fetches++;
    ns_retrieve_cfs_readonly
      (n, wrap(got_file_block, obj, gs, gl));
  }
  else if ((blocknum-CFS_NDIR) < SN_NFH) {
    n = rnode->indirect;
    obj->outstanding_fetches++;
    ns_retrieve_cfs_readonly
      (n, wrap (got_sindir_block, obj, blocknum - CFS_NDIR, gs, gl));
  }
  else if ((blocknum-CFS_NDIR-SN_NFH) < SN_NFH*SN_NFH) {
    n = rnode->double_indirect;
    uint32 dblock = (blocknum-CFS_NDIR-SN_NFH);
    obj->outstanding_fetches++;
    ns_retrieve_cfs_readonly
      (n, wrap (got_dindir_block, obj, dblock/SN_NFH, dblock%SN_NFH, gs, gl));
  }
  else if ((blocknum-CFS_NDIR-SN_NFH-SN_NFH*SN_NFH) < SN_NFH*SN_NFH*SN_NFH) {
    n = rnode->triple_indirect;
    uint32 tblock = (blocknum - CFS_NDIR - SN_NFH - SN_NFH*SN_NFH);
    uint32 toffst = tblock % (SN_NFH*SN_NFH);
    obj->outstanding_fetches++;
    ns_retrieve_cfs_readonly
      (n, wrap (got_tindir_block, obj, tblock/(SN_NFH*SN_NFH), toffst, gs, gl));
  }
  else
    warn << "fill_buf: read offset way too big! \n";
}

static void
read_got_file (ptr<read_obj> obj, ptr<cfs_data> data)
{
  if (!data || data->type != CFS_INODE) {
    obj->cb ((ptr<ns_block>) 0, true, CFSRWERR_IO);
    return;
  }
  
  cfs_inode *inode = data->inode;
  if (inode->type == CFSLNK) {
    cfs_inode_lnk *lnode = inode->lnk;
    memcpy (obj->buf, lnode->dest.cstr (), lnode->dest.len ());
    ptr<ns_block> b = New refcounted<ns_block> (obj->buf, lnode->dest.len ());
    obj->cb (b, true, CFSRW_OK);
  }
  else if (inode->type == CFSREG) {
    cfs_inode_reg *rnode = inode->reg;
    if (obj->rng.check(obj->offset, obj->count)) {
      //
      // got all the bytes, in here to check if we hit EOF or not
      //
      uint64 c = obj->rng.end () - obj->offset;
      if (c > obj->count)
	c = obj->count;
      bool eof = (obj->offset+c >= rnode->attr.size);
      ptr<ns_block> b = New refcounted<ns_block> (obj->buf, c);
      obj->cb (b, eof, CFSRW_OK);
    }
    else {
      //
      // we only want to read what was asked, so bytes
      // before obj->offset are done, and bytes after
      // obj->offset+obj->count, til end of the version of
      // the file stored in snapshot, are done. also, by
      // marking everything from obj->offset+obj->count to
      // eof as done, later on when we test for eof, we know
      // if we have hit eof or not.
      //
      if (obj->offset > 0)
        obj->rng.add (0, obj->offset);
      if ((obj->offset+obj->count) < rnode->attr.size)
        obj->rng.add
	  (obj->offset+obj->count,rnode->attr.size-(obj->offset+obj->count));
      //
      // if requested is more than file size, mark file size to
      // requested size as done. remember that a setattr delta
      // could have expanded the file size before scanning hit the
      // snapshot.
      //
      if ((obj->offset+obj->count) > rnode->attr.size &&
	  obj->offset < rnode->attr.size)
	obj->rng.add(rnode->attr.size, obj->offset+obj->count);
      obj->rng.gaps (wrap (fill_buf, obj, data));
      check_fill (obj);
    }
  }
}

void
cfsr::read (sptr s, bigint object, uint64 off, uint32 cnt,
            callback<void, ptr<ns_block>, bool, cfsrw_status>::ref cb)
{
  ptr<dbrec> q = bigint2dbrec (object);
  ptr<dbrec> v = obj2impl (s)->fmap_db->lookup (q);

  if (!v) {
    cb ((ptr<ns_block>) 0, true, CFSRWERR_IO);
    return;
  }

  ptr<read_obj> obj = New refcounted<read_obj> (obj2impl (s), cb);
  obj->offset = off;
  obj->count = cnt;
  obj->buf = New char[cnt];
  obj->len = 0;

  ns_retrieve_cfs_readonly
    (dbrec2bigint(v), wrap (read_got_file, obj));
}


//
// handle cfsr calls
//

static void
lookup_reply (cfsrcallp c, bigint object, cfsrw_status status)
{
  cfsr_lookup_res *res = static_cast<cfsr_lookup_res *> (c->res ());
  res->object = object;
  c->reply (status);
}

static void
getattr_reply (cfsrcallp c, ptr<cfsr_attr> attr, cfsrw_status status)
{
  cfsr_getattr_res *res = static_cast<cfsr_getattr_res *> (c->res ());
  res->attr = attr;
  c->reply (status);
}

static void
read_reply (cfsrcallp c, ptr<ns_block> data, bool eof, cfsrw_status status)
{
  cfsr_read_res *res = static_cast<cfsr_read_res *> (c->res ());
  res->data = data;
  res->eof = eof;
  c->reply (status);
}

static void
readdir_reply (cfsrcallp c, ptr<cfs_dirlist> dir, cfsrw_status status)
{
  cfsr_readdir_res *res = static_cast<cfsr_readdir_res *> (c->res ());
  res->dir = dir;
  c->reply (status);
}

static void
readlink_reply (cfsrcallp c, str dest, cfsrw_status status)
{
  cfsr_readlink_res *res = static_cast<cfsr_readlink_res *> (c->res ());
  res->dest = dest;
  c->reply (status);
}

static void
findver_reply (cfsrcallp c, cfsr_findver_res r, cfsrw_status status)
{
  cfsr_findver_res *res = static_cast<cfsr_findver_res *> (c->res ());
  res->logid = r.logid;
  res->cr_addr = r.cr_addr;
  c->reply (status);
}

void
cfsr::dispatch (cfsrcallp c)
{
  switch (c->proc ()) {
    case CFSRPROC_LOOKUP: 
      {
        cfsr_lookup_arg *arg = static_cast<cfsr_lookup_arg *> (c->arg ());
        cfsr::lookup (c->_s, arg->dir, arg->name, wrap (lookup_reply, c));
      }
      break;
    case CFSRPROC_GETATTR:
      {
        cfsr_getattr_arg *arg = static_cast<cfsr_getattr_arg *> (c->arg ());
        cfsr::getattr (c->_s, arg->object, wrap (getattr_reply, c));
      }
      break;
    case CFSRPROC_FINDVER:
      {
        cfsr_findver_arg *arg = static_cast<cfsr_findver_arg *> (c->arg ());
        cfsr::findver (c->_s, arg->object, arg->revision,
	               wrap (findver_reply, c));
      }
      break;
    case CFSRPROC_READ:
      {
        cfsr_read_arg *arg = static_cast<cfsr_read_arg *> (c->arg ());
        cfsr::read (c->_s, arg->file, arg->off, arg->cnt, wrap (read_reply, c));
      }
      break;
    case CFSRPROC_READLINK:
      {
        cfsr_readlink_arg *arg = static_cast<cfsr_readlink_arg *> (c->arg ());
        cfsr::readlink (c->_s, arg->file, wrap (readlink_reply, c));
      }
      break;
    case CFSRPROC_READDIR:
      {
        cfsr_readdir_arg *arg = static_cast<cfsr_readdir_arg *> (c->arg ());
        cfsr::readdir (c->_s, arg->dir, arg->cookie, arg->count,
	               wrap (readdir_reply, c));
      }
      break;
    default:
      assert (0);
      break;
  }
}


