
/*
 *  Copyright (C) 2002-2004  Massachusetts Institute of Technology
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 *  USA
 */

#include "replicaimpl.h"
#include "penviron.h"

void
replica_meta::serialize (strbuf &sb) const
{
  pkv.serialize (sb);
  commits.serialize (sb);
  suint (sb, leaves.size ());
  for (unsigned i=0; i<leaves.size (); i++)
    sbigint (sb, leaves [i]);
}

bool
replica_meta::parse (xin &x)
{
  if (!vervec::parse (&x, pkv))
    return false;
  if (!comvec::parse (&x, commits))
    return false;
  unsigned n;
  if (!x.ruint (n))
    return false;
  for (unsigned i=0; i<n; i++) {
    bigint a;
    if (!x.rbigint (a))
      return false;
    leaves.push_back (a);
  }
  return true;
}

replica_impl::replica_impl (bigint proj, bigint ra)
  : _projid (proj), _replicaaddr (ra)
{
  _mybranch = -1;
  _changed = false;
  
  _rdb = open_rdb ();
  _btree = 0;
  if (_rdb)
    _btree = New refcounted<btree_srv> (_rdb);
}

void
replica_impl::dispatch_with_snapshot (cfsrcallp c, sptr s,
                                      cfsrw_status status)
{
  if (status) {
    c->reply (CFSRWERR_NOREPLICA);
    return;
  }

  c->_s = s;
  cfsr::dispatch (c);
}

void
replica_impl::dispatch (cfsrcallp c)
{
  // Here, we finally start using the snapshot address to describe the
  // branch tree node because CFS needs the snapshot address.

  if (c->cr_addr == 0) {
    assert (_mybranch >= 0);
    c->_s = _branches [_mybranch];
    cfsr::dispatch (c);
  }
  else
    find (c->cr_addr, wrap (this, &replica_impl::dispatch_with_snapshot, c));
}

void 
replica_impl::sync (cbaddr cb)
{
  if (_branches.size ()) {
    assert (_rdb);
  
    strbuf b;
    _m.serialize (b);
    str s = b;
  
    ptr<dbrec> k = bigint2dbrec (_replicaaddr);
    ptr<dbrec> v = str2dbrec (s);
    _rdb->remove (k);
    _rdb->insert (k, v);
    _rdb->sync ();
 
    _changed = false;

    cb (_replicaaddr, CFSRW_OK);
  }
  else {
    cb (0, CFSRWERR_NOREPLICA);
  }
}

str
replica_impl::branch_tag (unsigned i) const
{
  assert (i < _branches.size ());
  return _branches [i]->btag ();
}

bigint
replica_impl::branch_head_snapshot (unsigned i) const
{
  assert (i < _branches.size ());
  return _branches [i]->addr ();
}

commit_info
replica_impl::branch_head_info (unsigned i) const
{
  assert (i < _branches.size ());
  return _branches [i]->branch ();
}

int
replica_impl::my_branch () const
{
  return _mybranch;
}

bool
replica_impl::set_branch_head (bigint n)
{
  unsigned index = get_branch_index(n);

  if (index < 0) 
    return false;
  else {
    _mybranch = index;
    return true;
  }
}

int 
replica_impl::get_branch_index (bigint n) const
{
  assert (_branches.size () > 0);
  for (unsigned i=0; i<_branches.size (); i++) {
    if (branch_head_info (i).rptr == n) {
      return i;
    }
  }
  return -1;
}

bigint
replica_impl::root () const
{
  assert (_branches.size () > 0);
  return _branches [0]->root ();
}

static inline void
sort_snapshots (spv& v)
{ 
  if (v.empty ())
    return;
  spv oldv = v;
  v.clear ();
  
  while (v.size () < oldv.size ()) {
    int x = -1;
    bigint h;
    for (unsigned i=0; i<oldv.size (); i++) {
      if (oldv [i] == 0)
        continue; 
      if (x == -1 || oldv [i]->addr () > h) {
        h = oldv [i]->addr ();
        x = i;
      }
    }
    assert (x >= 0);
    v.push_back (oldv [x]);
    oldv [x] = 0;
  }
}

void
replica_impl::set_addrs (spv current_branches)
{
  sort_snapshots (current_branches);
  int myb = -1;
  bool forked = false;

  // Figure out what the current branches are after all the new 
  // commit records are included.
  _branches.clear ();
  for (unsigned i=0; i<current_branches.size (); i++) {
    ptr<snapshot> s = obj2impl (current_branches [i]);
    if (s->flags () & commit_record::IGNORE_BRANCH)
      continue;
    if (_mybmagic != "" && s->ptag == _mybmagic) {
      if (myb == -1 && !forked)
        myb = i;
      else if (myb != -1) {
	//warn << "user specified branch forked\n";
	forked = true;
	myb = -1;
      }
    }
    bigint x = s->branch().rptr;
    _branches.push_back (s);
  }

  vec<bigint> v;
  for (unsigned i=0; i<_branches.size (); i++) {
    ptr<snapshot> si = obj2impl (_branches [i]);
    v.push_back (_branches [i]->addr ());
  }
  _m.leaves = v; // _m.leaves is a list of the branch leaf commit records

  if (myb >= 0)
    _mybranch = myb;
  else
    _mybranch = -1;
}

static inline void
add_snapshot_to_list (sptr s, spv& v)
{
  bigint na = s->addr ();
  bool found = false;
  for (unsigned i=0; i<v.size (); i++) {
    if (v [i]->addr () == na) {
      found = true;
      v [i] = s;
    }
  }
  if (!found)
    v.push_back (s);
}

// loads a set of snapshots as the latest snapshots for the
// repository. At the minimum a base snapshot is loaded (base = true).
// if there are branches, they are loaded as well.
// load_next gets called once for every log head.

void
replica_impl::load_next (vec<bigint> heads, cbcfsrw cb,
                         sptr s, cfsrw_status status)
{
  if (!status) {
    assert (s);
    ptr<snapshot> si = obj2impl (s);
    bigint x = s->branch().rptr;
    _branches.push_back (s);
    add_snapshot_to_list (s, _slist);

    if (!heads.empty ()) {
      bigint sn_addr = heads.pop_front ();
      load_snapshot (sn_addr, wrap (this, &replica_impl::load_next, heads, cb));
    }
    else {
      // loaded all snapshots
      cb (CFSRW_OK);
    }
  }
  else {
    _branches.clear ();
    cb (status);
  }
}

void
replica_impl::load (cbcfsrw cb)
{
  if (!_rdb) {
    cb (CFSRWERR_NOREPLICA);
    return;
  }

  ptr<dbrec> k = bigint2dbrec (_replicaaddr);
  ptr<dbrec> d = _rdb->lookup (k);
  if (!d) {
    cb (CFSRWERR_NOREPLICA);
    return;
  }

  str mx = dbrec2str (d);
  xin x (mx);
  if (!_m.parse (x)) {
    cb (CFSRWERR_NOREPLICA);
    return;
  }

  if (_m.leaves.size () == 0) {
    cb (CFSRWERR_NOREPLICA);
    return;
  }

  vec<bigint> log_heads = _m.leaves;
  bigint sn_addr = log_heads.pop_front();
  load_snapshot (sn_addr, wrap (this, &replica_impl::load_next, log_heads, cb));
}

ptr<pastdbfe>
replica_impl::open_rdb ()
{
  str r = pget_replicadir() << "/" << armor64A (_replicaaddr);
  str d = r << "/metadata.db";
  mkdir (r, 0755);

  dbOptions opts;
  opts.addOption("opt_flag", DB_CREATE);
  opts.addOption ("opt_cachesize", 1000);
  opts.addOption ("opt_nodesize", 4096);
  ptr<pastdbfe> db = New refcounted<pastdbfe> ();
  if (db->opendb (d.cstr (), opts))
    return 0;
  return db;
}

bool
replica_impl::lookup_snapshot (bigint c, bigint &s)
{
  // If the snapshot is *loaded*, then return it.  This will return
  // false even when the snapshot exists, but isn't loaded.

  if (!_btree){
    return false;
  }
  btree_node n;
  if (!_btree->lookup_by_commit (c, n))
    return false;

  s = crec_to_snapaddr (c);
  if (s == 0) 
    return false;
  return true;
}

// find a particular snapshot using its address. could return a cached
// snapshot object, or load the snapshot if it does not exists.

void
replica_impl::load_cb (cbloadsn_t cb, sptr s, cfsrw_status status)
{
  if (!status) {
    assert (s);
    add_snapshot_to_list (s, _slist);
  }
  cb (s, status);
}

static void
find_cb (cbloadsn_t cb, sptr s)
{
  cb (s, CFSRW_OK);
}


void
replica_impl::find (bigint cr_addr, cbloadsn_t cb)
{
  for (unsigned i=0; i<_slist.size (); i++) {
    if (_slist [i]->branch().rptr == cr_addr) {
      delaycb (0, wrap (find_cb, cb, _slist[i]));
      return;
    }
  }
  bigint sn_addr = crec_to_snapaddr(cr_addr);
  load_snapshot
    (sn_addr, wrap (this, &replica_impl::load_cb, cb));
}

// builds snapshots, either from scratch, if no base snapshot is
// loaded, or from previously built snapshots.

void
replica_impl::build_cb (comvec commits, cbcfsrw cb, spv v, cfsrw_status status)
{
  if (status && status != CFSRWERR_NONEWRECS) {
    warn << "cannot build snapshot, err " << status << "\n";
    _branches.clear ();
    _mybranch = -1;
    cb (status);
    return;
  }

  _changed = true;
  nsobj->sync ();

  if (status == CFSRWERR_NONEWRECS) {
    //warn << "no new deltas, no new snapshots built\n";
    _m.commits = commits;
    v.clear ();
    v = _branches;
  }
  else {
    assert (v.size () > 0);
    for (unsigned j=0; j<v.size (); j++)
      add_snapshot_to_list (v[j], _slist);
    _m.commits = commits;
  }

  set_addrs (v);
  cb (CFSRW_OK);
}

void
replica_impl::update (comvec commits, cbcfsrw cb)
{
  if (!_rdb) {
    cb (CFSRWERR_NOREPLICA);
    return;
  }

  if (_mybranch >= 0) {
    char xxb[20];
    rnd.getbytes (xxb, 20);
    _mybmagic = str (xxb, 20);
    ptr<snapshot_obj> s = _branches [_mybranch];
    _mybptag = obj2impl (s)->ptag;
    obj2impl (s)->ptag = _mybmagic;
  }
  else
    _mybmagic = "";

  if (_branches.size () > 0) {
    build_snapshot
      (this, _projid, commits,
       wrap (this, &replica_impl::build_cb, commits, cb), _branches);
  } else {
    build_snapshot
      (this, _projid, commits,
       wrap (this, &replica_impl::build_cb, commits, cb));
  }
}

bigint
replica_impl::determine_current_branch(str arg_branch, bool ignore_implicit_branches, 
				       str crec_from_info) const
{
  // Find and set current branch
  // -- Checks the various forms that arg_branch can take.
  // if we can't find a branch with arg_branch, then try the
  // commit_record from the info file if the user is ignoring implicit
  // forks.
  bigint c=0;
  str current_guess = arg_branch;

  // Compare current_guess to current branch heads.
  if (current_guess && current_guess != "") {
    str n;
    vec<bigint> matching_branch_heads;

    for (unsigned i=0; i<nbranches (); i++) {
      if (current_guess == branch_tag (i)) {
	c = branch_head_info (i).rptr;
	matching_branch_heads.push_back(c);
      }
    }

    if (matching_branch_heads.size() == 1) {
      return matching_branch_heads[0];

    } else if (matching_branch_heads.size() > 1 && ignore_implicit_branches && 
	       crec_from_info && crec_from_info != "") {
      // read commit_record addr from info file and search for a branch head
      // that is a sole descendant of the commit_record.
      bigint cr;
      bigint head;
      //str commit_rec = get_commit_record_from_infofile(file);
      if (dearmor64A_bigint (crec_from_info, cr)) {
	if (_btree->get_sole_descendant (cr, head)) {

	  // Check that the branchtags match.
	  str headtag = branch_tag (get_branch_index(head));
	  if (headtag != current_guess) {
	    warn << "Chose head " << armor64A(head) << "(" << headtag 
		 << ") because of commit record " << armor64A(cr) 
		 << " in Pastwatch/info, but branch tag \"" 
		 << current_guess << "\" does not match.\n";
	    return 0;
	  }
	  return head;
	}
      }
    } else if (matching_branch_heads.size() > 1) {
      warn << "Branch \"" << current_guess << "\" has forked.  You must specify a "
	   << "branch head to track with the -b option.\n";
      return 0;
    }
  }
  
  // Compare current_guess to tagged versions
  if (current_guess && current_guess != "") {
    if (_btree->lookup_by_tag(current_guess, c)){
      return c;
    }
  }

  // Check if there is only one branch
  if (!current_guess || current_guess == "") {
    if (nbranches () == 1) {
      return branch_head_info (0).rptr;
    }
  }

  // Check if current_guess is equal to the branch's commit_record key.
  if (current_guess && current_guess != "" && 
      dearmor64A_bigint (current_guess, c)){
    return c;
  }

  return c;
}

int 
replica_impl::map_crec_to_snapaddr (bigint crecord, bigint snapshot)
{
  /*   Blocks in db map commit record addresses to their
   *   snapshot addresses.  { cr:enMD832OsfQ2AP61B2K0hWxWNSI -> nIuUVNE97PPyRfdGm3Frl7vFIBc }
   */

  strbuf sb;
  sb << "cr:" << armor64A (crecord);
  ptr<dbrec> key = str2dbrec (str(sb));
  ptr<dbrec> new_value = bigint2dbrec (snapshot);

  ptr<dbrec> current_value = _rdb->lookup (key);
  if (current_value)
    _rdb->remove (key);

  return _rdb->insert (key, new_value);
}

bigint
replica_impl::crec_to_snapaddr (bigint crecord)
{
  strbuf sb;
  sb << "cr:" << armor64A (crecord);
  ptr<dbrec> key = str2dbrec (str(sb));
  ptr<dbrec> value = _rdb->lookup (key);
  bigint saddr = (value) ? dbrec2bigint(value) : 0;

  if (saddr == 0) 
    warn << "Could not find snapshot for commit record " 
	 << armor64A (crecord) << ".  Consider rebuilding "
	 << "the local snapshots by doing a fresh past checkout.\n";
  return saddr;
}
