/*
 *  Copyright (C) 2002-2004  Massachusetts Institute of Technology
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 *  USA
 */

#include "stdinc.h"
#include "pastdbfe.h"
#include "build.h"
#include "snapshot.h"
#include "replicaimpl.h"

#define FMAP_ENTRIES_PER_PAGE SN_NFH

void
bob::store_new (snode *e)
{
  if (e->ent->type == CFS_INODE)
    to_store.push_back (e->inum);
}

void
bob::store_tree ()
{
  if (sobj->thisinfo.prevaddr == 0) {
    snode *se = sobj->active_snodes [sobj->thisinfo.rootdir];
    if (!se) {
      warn << "no information on root directory\n";
      scb ((ptr<snapshot>) 0, CFSRWERR_ERROR);
      return;
    }
    se->ent->inode->reg->attr.nlink++; // rootdir nlink is always 1
    store (sobj->thisinfo.rootdir);
  }
  else if (sobj->active_snodes.size() > 0) {
    to_store.clear ();
    sobj->active_snodes.traverse 
      (wrap (mkref (this), &bob::store_new));
    if (outstanding_inserts == 0)
      insert_snapshot ();
  }
  else
    insert_snapshot();
}

void 
bob::insert_snapshot ()
{
  if (to_store.size ()) {
    while (to_store.size ()) {
      bigint n = to_store.pop_front ();
      store (n);
    }
    if (outstanding_inserts == 0)
      insert_snapshot ();
  }

  else if (history_update_rpcs == 0) {
    assert (outstanding_inserts == 0);

    if (insert_failures) {
      warn << "there are insert failures\n";
      scb ((ptr<snapshot>)0, CFSRWERR_ERROR);
    }
    else
      serialize_and_store ();
  }
}

void 
bob::insert_cb (bool map, bigint inum, ns_stat status, bigint key)
{
  outstanding_inserts--;

  if (status == NS_OK) {
    if (map) {
      // insert inum->chash to fmap
      ptr<dbrec> k = bigint2dbrec (inum);
      ptr<dbrec> d = bigint2dbrec (key);
      assert (sobj->fmap_db);
      if (sobj->fmap_db->insert (k, d))
	insert_failures++;
    }
  }
  else {
    warn << "bob::insert_cb: dhash insert failed.\n";
    insert_failures++;
  }

  if (outstanding_inserts == 0)
    insert_snapshot ();
}

bigint
bob::insert (char *data, size_t len)
{
  outstanding_inserts++;
  inserts++;
  bigint n = nsobj->insert_local (data, len);
  delaycb (0, wrap (mkref (this), &bob::insert_cb, false, 0, NS_OK, n));
  xfree (data);
  return n;
}

bigint
bob::insert_and_addtofmap (char *data, size_t len, bigint inum)
{
  outstanding_inserts++;
  inserts++;
  bigint n = nsobj->insert_local (data, len);
  delaycb (0, wrap (mkref (this), &bob::insert_cb, true, inum, NS_OK, n));
  xfree (data);
  return n;
}

bigint 
bob::store (bigint n) 
{
  snode *e = sobj->active_snodes [n];
  assert (e && e->ent->type == CFS_INODE);
  
  cfs_ftype type = e->ent->inode->type;
  cfs_inode *inode = e->ent->inode;
  if (type == CFSLNK)
    return store_inode (n, inode);
  else if (type == CFSREG)
    return store_file (n, inode);
  else
    return store_directory (n, inode);
}

bigint
bob::store_inode (bigint inum, cfs_inode *inode) 
{
  ptr<cfs_data> dat = New refcounted<cfs_data> (CFS_INODE);
  *(dat->inode) = *inode;

  size_t xlen = 0;
  char *xbuf = NULL;
  xdrsuio x (XDR_ENCODE);
  if (xdr_cfs_data (x.xdrp (), dat)) {
    xlen = x.uio ()->resid ();
    xbuf = suio_flatten (x.uio ());
  }
  return insert_and_addtofmap (xbuf, xlen, inum);
}

bigint 
bob::store_directory (bigint inum, cfs_inode *inode)
{
  unsigned blk = 0;
  unsigned dirsize = 0;

  for (blk = 0; blk < inode->reg->direct.size(); blk++) {
    bigint n = inode->reg->direct[blk];
    snode *db = sobj->active_snodes [n];
    if (!db) // dirblk not modified or loaded
      continue;
    assert (db->ent->type == CFS_DIRBLK);

    cfs_directory *dir = db->ent->dir;
    cfs_dirent *dirent = dir->entries;
    
    while (dirent) {
      n = dirent->fh;
      // in build from prev, already storing everything in
      // active_snodes
      if (sobj->thisinfo.prevaddr == 0)
        store (n);
      dirent = dirent->nextentry;
    }
  
    size_t xlen = 0;
    char *xbuf = NULL;
    xdrsuio x (XDR_ENCODE);
  
    if (xdr_cfs_data (x.xdrp (), db->ent)) {
      xlen = x.uio ()->resid ();
      xbuf = suio_flatten (x.uio ());
    }
    n = insert (xbuf, xlen);
    dirsize += xlen;
    inode->reg->direct[blk] = n;
  }
  // should be size of all blocks, not just those loaded
  inode->reg->attr.size = dirsize;

  return store_inode (inum, inode);
}

bigint 
bob::store_file (bigint inum, cfs_inode *inode)
{
  return store_inode (inum, inode);
}
 
bigint
bob::store_one_fmap_block (str s)
{
  fmap_block_inserts++;
  inserts++;
  bigint n = nsobj->insert_local (s.cstr (), s.len ());
  delaycb (0, wrap (mkref (this), &bob::insert_fmapblock_cb, NS_OK, n));
  return n;
}

void
bob::serialize_and_store ()
{
  fmap_block_inserts = 0;
  fmap_block_insert_failures = 0;

  vec<bigint> inum;
  vec<bigint> chash;
  ptr<pdbEnumeration> it = sobj->fmap_db->enumerate ();
  ptr<dbrec> k = it->nextElement ();
  while (k) {
    ptr<dbrec> d = sobj->fmap_db->lookup (k);
    inum.push_back (dbrec2bigint (k));
    chash.push_back (dbrec2bigint (d));
    k = it->nextElement ();
  }
  //it = NULL;

  bigint next = 0;
  unsigned total = 0;
  while (inum.size () > FMAP_ENTRIES_PER_PAGE) {
    unsigned n=0;
    strbuf sb;
    sint (sb, FMAP_ENTRIES_PER_PAGE);  // entries in this block
    sint (sb, total);                  // remaining entries after this block
    sbigint  (sb, next);                   // address of next block
    while (n < FMAP_ENTRIES_PER_PAGE) {
      bigint i = inum.pop_back ();
      bigint c = chash.pop_back ();
      sbigint (sb, i);
      sbigint (sb, c);
      total++;
      n++;
    }
    next = store_one_fmap_block (sb);
  }

  strbuf sb;
  sstr (sb, VERSION);
  sobj->thisinfo.branch.vptr = 0;
  sobj->thisinfo.serialize (sb);

  sint (sb, inum.size ()); // entries in this block
  sint (sb, total);        // remaining entries after this block
  sbigint (sb, next);      // address of next block
  while (inum.size () > 0) {
    bigint i = inum.pop_back ();
    bigint c = chash.pop_back ();
    sbigint (sb, i);
    sbigint (sb, c);
  }

  sobj->thisaddr = store_one_fmap_block (sb);
  sobj->thisinfo.branch.vptr = sobj->thisaddr;
  _replica->map_crec_to_snapaddr(commit.rptr, sobj->thisaddr);

  // warn << "snapshot: " << inserts << " cfs blocks inserted\n";
  bigint cr = commit.rptr;
  warn << "new snapshot: "
       << armor64A (sobj->thisaddr) << ", " << ndeltas << " deltas.  "
       << "For commit record: " << armor64A(cr) << "\n";
}

void 
bob::insert_fmapblock_cb (ns_stat status, bigint key)
{
  assert (fmap_block_inserts > 0);
  fmap_block_inserts--;
  if (status != NS_OK)
    fmap_block_insert_failures++;

  if (fmap_block_inserts == 0) {
    if (!fmap_block_insert_failures)
      scb (sobj, CFSRW_OK);
    else {
      warn << "snapshot: dhash insert failed.\n";
      scb ((ptr<snapshot>) 0, CFSRWERR_ERROR);
    }
  }
}

