
#include "dhash_common.h" // keyhash_payload
#include "dhash.h"
#include "verify.h"
#include <dbfe.h>
#include <sfscrypt.h>
#include <dhash_prot.h>
#include <chord.h>
#include <chord_types.h>
#include <id_utils.h>

bigint
compute_hash (const void *buf, size_t buflen)
{
  char h[sha1::hashsize];
  bzero(h, sha1::hashsize);
  sha1_hash (h, buf, buflen);
  
  bigint n;
  mpz_set_rawmag_be(&n, h, sha1::hashsize);
  return n;
}

bool
verify (chordID key, dhash_ctype t, const char *buf, int len) 
{
  switch (t) {
  case DHASH_CONTENTHASH:
    return verify_content_hash (key, buf, len);
    break;
  case DHASH_KEYHASH:
    return verify_keyhash (key, buf, len);
    break;
  case DHASH_NOAUTH:
  case DHASH_APPEND:
    return true;
    break;
  default:
    warn << "bad type " << t << "\n";
    return false;
  }
  return false;
}

bool
verify_content_hash (chordID key, const char *buf, int len) 
{
  char hashbytes[sha1::hashsize];
  sha1_hash (hashbytes, buf, len);
  chordID ID;
  mpz_set_rawmag_be (&ID, hashbytes, sizeof (hashbytes));  // For big endian
  return (ID == key);
}

bool
verify_keyhash (chordID key, const char *buf, int len)
{
  // extract the public key from the message
  sfs_pubkey2 pubkey;
  sfs_sig2 sig;
  long plen;

  xdrmem x (buf, (unsigned)len, XDR_DECODE);
  if (!xdr_sfs_pubkey2 (&x, &pubkey))
    return false;
  if (!xdr_sfs_sig2 (&x, &sig))
    return false;
  if (!XDR_GETLONG (&x, &plen))
    return false;
  if (plen <= 0)
    return false;

  ptr<keyhash_payload> p = keyhash_payload::decode (x, plen);
  if (!p) {
    warn << "cannot decode payload\n";
    return false;
  }

  chordID hash = p->id (pubkey);
  if (hash != key) {
    warn << "key mismatch\n";
    return false;
  }

  // verify signature
  bool ok = p->verify (pubkey, sig);
  if (!ok)
    warn << "sig mismatch\n";
  return ok;
}

ptr<dhash_block>
get_block_contents (ptr<dbrec> d, dhash_ctype t)
{
  return get_block_contents (d->value, d->len, t);
}

ptr<dhash_block>
get_block_contents (ref<dbrec> d, dhash_ctype t)
{
  return get_block_contents (d->value, d->len, t);
}


ptr<dhash_block> 
get_block_contents (ptr<dhash_block> block, dhash_ctype t) 
{
  return get_block_contents (block->data, block->len, t);
}

ptr<dhash_block> 
get_block_contents (const char *data, unsigned int len, dhash_ctype t)
{
  char *content;
  long contentlen = len;

  xdrmem x (data, len, XDR_DECODE);
  switch (t) {
  case DHASH_KEYHASH:
    {
      sfs_pubkey2 k;
      sfs_sig2 s;
      if (!xdr_sfs_pubkey2 (&x, &k) ||
	  !xdr_sfs_sig2 (&x, &s) ||
	  !XDR_GETLONG (&x, &contentlen))
	return NULL;
    }
    /* FALL THROUGH */

  case DHASH_CONTENTHASH:
  case DHASH_NOAUTH:
  case DHASH_APPEND:
    {
      if (contentlen >= 0 && !(content = (char *)XDR_INLINE (&x, contentlen)))
	return NULL;
    }
    break;

  default:
    return NULL;
  }

  ptr<dhash_block> d = New refcounted<dhash_block> (content, contentlen, t);
  return d;
}

long
keyhash_version (const char *value, unsigned int len)
{
  xdrmem x (value, len, XDR_DECODE);
  sfs_pubkey2 k;
  sfs_sig2 s;
  long plen;

  if (!xdr_sfs_pubkey2 (&x, &k) ||
      !xdr_sfs_sig2 (&x, &s) ||
      !XDR_GETLONG (&x, &plen))
    return -1;
  if (plen <= 0)
    return -1;

  ptr<keyhash_payload> p = keyhash_payload::decode (x, plen);
  if (!p)
    return -1;
  return p->version ();
}

long
keyhash_version (ptr<dhash_block> data)
{
  return keyhash_version (data->data, data->len);
}

long
keyhash_version (ref<dhash_block> data)
{
  return keyhash_version (data->data, data->len);
}

long
keyhash_version (ref<dbrec> data)
{
  return keyhash_version (data->value, data->len);
}

long
keyhash_version (ptr<dbrec> data)
{
  return keyhash_version (data->value, data->len);
}

int
keyhash_projkey (const char *value, unsigned int len, bigint& n)
{
  xdrmem x (value, len, XDR_DECODE);
  sfs_pubkey2 k;
  sfs_sig2 s;
  long plen;

  if (!xdr_sfs_pubkey2 (&x, &k) ||
      !xdr_sfs_sig2 (&x, &s) ||
      !XDR_GETLONG (&x, &plen))
    return -1;
  if (plen <= 0)
    return -1;

  ptr<keyhash_payload> p = keyhash_payload::decode (x, plen);
  if (!p)
    return -1;
  
  n = p->projkey ();
  return 0;
}

int
keyhash_projkey (ptr<dhash_block> data, bigint& k)
{
  return keyhash_projkey (data->data, data->len, k);
}

int
keyhash_projkey (ref<dhash_block> data, bigint& k)
{
  return keyhash_projkey (data->data, data->len, k);
}

int
keyhash_projkey (ref<dbrec> data, bigint& k)
{
  return keyhash_projkey (data->value, data->len, k);
}

int
keyhash_projkey (ptr<dbrec> data, bigint& k)
{
  return keyhash_projkey (data->value, data->len, k);
}

ptr<sfs_pubkey2>
keyhash_pk (const char *value, unsigned int len)
{
  xdrmem x (value, len, XDR_DECODE);
  ptr<sfs_pubkey2> k = New refcounted<sfs_pubkey2> ();

  if (!xdr_sfs_pubkey2 (&x, k)) 
    return NULL;
  return k;
}


/*
 * keyhash_payload
 */

keyhash_payload::keyhash_payload (bigint projkey, long version, str buf)
  : _projkey (projkey), _version (version), _buf (buf)
{
  memset (_salt, 0, sizeof (salt_t));
}

keyhash_payload::keyhash_payload (bigint projkey, const salt_t s,
                                  long version, str buf)
  : _projkey (projkey), _version (version), _buf (buf)
{
  memmove (_salt, s, sizeof (salt_t));
}

// empty payload, useful only for computing ids
keyhash_payload::keyhash_payload ()
  : _projkey (0), _version (0), _buf ("")
{
  memset (_salt, 0, sizeof (salt_t));
}

keyhash_payload::keyhash_payload (bigint projkey, const salt_t s)
  : _projkey (projkey), _version (0), _buf ("")
{
  memmove (_salt, s, sizeof (salt_t));
}

chordID
keyhash_payload::id (sfs_pubkey2 pk) const
{
  char digest[sha1::hashsize];

  // get the public key into raw form
  strbuf pkstrbuf;
  ptr<sfspub> pubkey = sfscrypt.alloc (pk, SFS_VERIFY);
  pubkey->export_pubkey (pkstrbuf, false);
  str pubkeystr(pkstrbuf);

  sha1ctx ctx;
  ctx.update (pubkeystr.cstr(), pubkeystr.len());
  ctx.update (_salt, sizeof (salt_t));
  ctx.final (digest);

  bigint chordid;
  mpz_set_rawmag_be(&chordid, digest, sha1::hashsize);
  return chordid;
}

// the signature is taken over:
//    salt_t   (20 opaque bytes)
//    projkey  (20 opaque bytes)
//    version  (4 bytes, in network byte order)
//    user_payload
//
//    To sign, we use a string constructed with an iovector to concatenate 
//    these elements and then we sign the string. On the verify side, we
//    again form a string out of what we think the three elements should be
//    (we get them from the flattened xdr struct that is stored inside
//    dhash), and then we verify this new string with the public key (also 
//    stored inside the xdr structure)
void
keyhash_payload::sign (ptr<sfspriv> key, sfs_pubkey2& pk, sfs_sig2& sig) const
{
  long version_nbo = htonl (_version);
  char b [sha1::hashsize];
  memset (b, 0, sha1::hashsize);
  mpz_get_rawmag_be (b, sha1::hashsize, &_projkey);

  iovec iovs [4];
  iovs [0].iov_base = const_cast<char *> (_salt);
  iovs [0].iov_len = sizeof (salt_t);
  iovs [1].iov_base = const_cast<char *> (b);
  iovs [1].iov_len = sha1::hashsize;
  iovs [2].iov_base = reinterpret_cast<char *> (&version_nbo);
  iovs [2].iov_len = sizeof (long);
  iovs [3].iov_base = const_cast<char *> (_buf.cstr ());
  iovs [3].iov_len = _buf.len ();
  str m (iovs, 4);
  key->sign (&sig, m);
  key->export_pubkey (&pk); 
}
  
bool
keyhash_payload::verify (sfs_pubkey2 pubkey, sfs_sig2 sig) const
{
  long version_nbo = htonl (_version);
  char b [sha1::hashsize];
  memset (b, 0, sha1::hashsize);
  mpz_get_rawmag_be (b, sha1::hashsize, &_projkey);

  ptr<sfspub> pk = sfscrypt.alloc (pubkey, SFS_VERIFY);
  iovec iovs [4];
  iovs [0].iov_base = const_cast<char *> (_salt);
  iovs [0].iov_len = sizeof (salt_t);
  iovs [1].iov_base = const_cast<char *> (b);
  iovs [1].iov_len = sha1::hashsize;
  iovs [2].iov_base = reinterpret_cast<char*> (&version_nbo);
  iovs [2].iov_len = sizeof (long);
  iovs [3].iov_base = const_cast<char *> (_buf.cstr ());
  iovs [3].iov_len = _buf.len ();
  str m (iovs, 4);
  return pk->verify (sig, m);
}

int
keyhash_payload::encode (xdrsuio &x) const
{
  long v = _version;

  if (!XDR_PUTLONG (&x, &v))
    return -1;

  // stuff the salt into the xdr structure
  char* salt_slot;
  if ((salt_slot = (char *)XDR_INLINE(&x, sizeof (salt_t))) == NULL)
    return -1;
  memmove (salt_slot, _salt, sizeof (salt_t));
  
  char* projkey_slot;
  if ((projkey_slot = (char *)XDR_INLINE(&x, sha1::hashsize)) == NULL)
    return -1;
  mpz_get_rawmag_be (projkey_slot, sha1::hashsize, &_projkey);

  char *m_buf;
  int size = _buf.len () + 3 & ~3;
  if (!(m_buf = (char *)XDR_INLINE (&x, size)))
    return -1;
  memcpy (m_buf, _buf.cstr (), _buf.len ());
  return 0;
}

ptr<keyhash_payload>
keyhash_payload::decode (xdrmem &x, long plen)
{
  ptr<keyhash_payload> p = 0;
  long version;
  char *salt;
  char *pk;
  char *buf;
  bigint projkey;

  if (plen < (long)((sizeof (long))+sha1::hashsize+sizeof (salt_t)))
    return p;

  if (!XDR_GETLONG (&x, &version))
    return p;
  if (!(salt = (char *)XDR_INLINE (&x, sizeof (salt_t))))
    return p;
  if (!(pk = (char *)XDR_INLINE (&x, sha1::hashsize)))
    return p;
  mpz_set_rawmag_be (&projkey, pk, sha1::hashsize);

  long buflen = plen - sizeof (long) - sha1::hashsize - sizeof (salt_t);
  if (!(buf = (char *)XDR_INLINE (&x, buflen)))
    return p;

  p = New refcounted<keyhash_payload>
    (projkey, salt, version, str (buf, buflen));
  return p;
}

ptr<keyhash_payload>
keyhash_payload::decode (ptr<dhash_block> b)
{
  xdrmem x (b->data, b->len, XDR_DECODE);
  return decode (x, b->len);
}

bool
sig_hashlist::contains(bigint hash) {
  u_int16_t i;
  for(i=0; i<hashes.size(); i++) {
    if (hashes[i] == hash)
      return true;
  }
  return false;
}

int 
sig_hashlist::insert(bigint hash) {
  if (contains(hash)) return 0;
  hashes.push_back(hash);
  return 1;
}

int
sig_hashlist::encode(xdrsuio &x) {
  // encoded block contains:
  // 1 number of bytes of hashes (long)
  // 2 hashes separated by commas
  // 3 receipt (xdr)
  str all_hashes(get_all_hashes());
  long len = all_hashes.len();

  if (!XDR_PUTLONG (&x, &len))
    return 0;
  
  char *m_buf;
  int size = (all_hashes.len () + 3) & ~3;
  
  if (!(m_buf = (char *)XDR_INLINE (&x, size)))
    return 0;
  memcpy (m_buf, all_hashes.cstr (), len);

  if (!xdr_sfs_sig2 (&x, &sig))
    return 0;
  return 1;
}

int
sig_hashlist::decode(xdrmem &x) {
  long hash_len;
  const char *m_buf, *p, *front;
  long size;

  if (!XDR_GETLONG (&x, &hash_len))
    return 0;

  size = (hash_len + 3) & ~3;
  if (!(m_buf = (char *)XDR_INLINE (x.xdrp(), size)))
    return 0;
 
  str all_hashes(m_buf, hash_len);
  front = m_buf = all_hashes.cstr();
  while ((front < m_buf + strlen(m_buf)) && (p = strchr(front, ','))) {
    hashes.push_back( bigint(str(front, p-front)));
    front = p+1;
  }

  if (!xdr_sfs_sig2 (&x, &sig))
    return 0;

  return 1;
}

str 
sig_hashlist::get_all_hashes() {
  u_int16_t i;
  strbuf all_hashes;
  for(i=0; i<hashes.size(); i++) {
    all_hashes << "0x" << hashes[i] << ",";
  }
  return all_hashes;
}

void 
secure_receipt::sign(ptr<sfspriv> signer) {
  signer->sign(&sig, get_all_hashes());
}

bool
secure_receipt::verify(ptr<sfs_pubkey2> pubkey) {
  ptr<sfspub> pk = sfscrypt.alloc (*pubkey, SFS_VERIFY);
  strbuf buf(get_all_hashes());
  return pk->verify(sig, str(buf));
}

