/*
 *  Copyright (C) 2002-2004  Massachusetts Institute of Technology
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 *  USA
 */

#include <errno.h>
#include "pastwatch.h"
#include "opts.h"
#include "pdb.h"
#include "path.h"
#include "io.h"

extern vec<str> notify_cmds;
extern vec<str> notify_files;

class rename_object : public virtual refcount
{
private:
  bigint root;
  str from;
  str to;
  cbi cb;
  str cfsfrom;
  str cfsto;

  bigint cr_addr;
  mtime_t mtime;

  bigint object;
  bigint fromdir;
  bigint todir;
  str fromname;
  cfsw_ftype ftype;

  void
  rename_reply (cfswcallp cc, cfsrw_status status)
  {
    if (status) {
      warn << "cannot rename " << cfsfrom << " to " << cfsto << "\n";
      cb (-1);
      return;
    }
    
    if (rename (from, to)) {
      warn << "cannot rename " << from << "to "
	   << to << " in working directory\n";
      cb (-1);
      return;
    }

    pastwatch_status ("Renaming", cfsfrom, cfsto);
    // for commit notification message
    notify_cmds.push_back ("A");
    notify_files.push_back (cfsto);
      
    pmgr->schedule_add (get_parent_dir (from), pfe (fromdir));
    pmgr->schedule_add (get_parent_dir (to), pfe (todir));
    pmgr->schedule_add (to, pfe (object, cr_addr, mtime));
    pmgr->schedule_remove (from);
    if (ftype == CFSW_DIR)
      pmgr->schedule_rename (to, cfsto);

    cb (0);
  }

  void
  lookup_to_reply (cfsrcallp cc, cfsrw_status status)
  {
    if (!status) {
      warn << cfsto << " exists in repository, remove first\n";
      cb (-1);
      return;
    }
    if (status && status != CFSRWERR_NOENT) {
      warn << "cannot access repository to lookup " << cfsto << "\n";
      cb (-1);
      return;
    }

    cfsr_lookup_arg *a = static_cast<cfsr_lookup_arg *> (cc->arg ());
    todir = a->dir;
    str toname = a->name;

    cfswcallp c = New refcounted<cfswcall>
      (CFSWPROC_RENAME, wrap (mkref (this), &rename_object::rename_reply));
    cfsw_rename_arg *arg = static_cast<cfsw_rename_arg *> (c->arg ());
    arg->object = object;
    arg->type = ftype;
    arg->from_dir = fromdir;
    arg->from_name = fromname;
    arg->to_dir = todir;
    arg->to_name = toname;
    arg->target_exists = false;
    cfsw_call (c);
  }

  void
  getattr_reply (cfsrcallp cc, cfsrw_status status)
  {
    if (status) {
      warn << "cannot get attribute of " << cfsfrom << " in repository\n";
      cb (-1);
      return;
    }
    cfsr_getattr_res *res = static_cast<cfsr_getattr_res *> (cc->res ());
    if (res->attr->type == CFSREG)
      ftype = CFSW_REG;
    else if (res->attr->type == CFSDIR)
      ftype = CFSW_DIR;
    else if (res->attr->type == CFSLNK)
      ftype = CFSW_LNK;
    else
      assert (0);
    cfsr_lookup
      (root, cfsto, wrap (mkref (this), &rename_object::lookup_to_reply));
  }

  void
  lookup_from_reply (cfsrcallp cc, cfsrw_status status)
  {
    if (status) {
      warn << "cannot lookup " << cfsfrom << " in repository\n";
      warn << "already renamed?\n";
      cb (-1);
      return;
    }
    cfsr_lookup_arg *a = static_cast<cfsr_lookup_arg *> (cc->arg ());
    cfsr_lookup_res *res = static_cast<cfsr_lookup_res *> (cc->res ());

    ptr<pfe> x = pmgr->lookup (from);
    assert (x);
    if (x->fh != res->object) {
      warn << cfsfrom << " is a different file in the current repository!\n";
      cb (-1);
      return;
    }

    cr_addr = x->cr_addr;
    mtime = x->mtime;

    fromdir = a->dir;
    fromname = a->name;
    object = res->object;
        
    cfsrcallp c = New refcounted<cfsrcall>
      (CFSRPROC_GETATTR, wrap (mkref (this), &rename_object::getattr_reply));
    cfsr_getattr_arg *arg = static_cast<cfsr_getattr_arg *> (c->arg ());
    arg->object = object;
    cfsr_call (c);
  }

public:
  rename_object (bigint root, str fromp, str top, cbi cb)
    : root (root), from (fromp), to (top), cb (cb)
  {
    struct stat sb;
    int err = lstat (to, &sb);
    if (!err && (sb.st_mode & S_IFDIR) == S_IFDIR) {
      vec<str> x;
      splitpath (x, from);
      assert (x.size () > 0);
      to = to << "/" << x [x.size ()-1];
      to = sanitize_path (to);

      err = lstat (to, &sb);
      if (!err || errno != ENOENT) {
        warn << "cannot stat(4) " << to << ", or file exists\n";
        cb (-1);
        return;
      }
    }
      
    if (!err || errno != ENOENT) {
      warn << "cannot stat(5) " << to << ", or file exists\n";
      cb (-1);
      return;
    }
    err = lstat (from, &sb);
    if (err) {
      warn << "cannot stat(6) " << from << "\n";
      cb (-1);
      return;
    }

    if (get_cfspath (from, cfsfrom)) {
      warn << "cannot get path of " << from << " in repository\n";
      cb (-1);
      return;
    }
    if (get_cfspath (to, cfsto)) {
      warn << "cannot get path of " << to << " in repository\n";
      cb (-1);
      return;
    }
   
    if (!pmgr->lookup (from)) {
      warn << "cannot lookup " << from << " in local database\n";
      cb (-1);
      return;
    }

    if (pmgr->lookup (to)) {
      warn << to << " already in local database\n";
      cb (-1);
      return;
    }

    // verify in the latest snapshot of the repository
    cfsr_lookup
      (root, cfsfrom, wrap (mkref (this), &rename_object::lookup_from_reply));
  }
};

void
rename (bigint root, str from, str to, cbi cb)
{
  ptr<rename_object> a = New refcounted<rename_object> (root, from, to, cb);
}

